/*
 * s_midi.h
 *
 */

//
// Programmer:    Craig Stuart Sapp [craig@ccrma.stanford.edu]
// Creation Date: Mon Dec 21 18:00:42 PST 1998
// Last Modified: Mon Dec 21 18:00:42 PST 1998
// Filename:      ...linuxmidi/output/method1.c
// Syntax:        C 
// $Smake:        gcc -O -o devmidiout devmidiout.c && strip devmidiout
//

/*
 *	Update: 2016/05/04 by kurino@math.cst.nihon-u.ac.jp
 *	Update: 2017/04/28 by kurino@math.cst.nihon-u.ac.jp
 *		s_midi_play_speed を追加
 */

#ifndef	__S_MIDI_H__
#define	__S_MIDI_H__

/*
 *
 */

#include <linux/soundcard.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>

#include <ctype.h>
#include <string.h>

/*
 *
 */

#define	EOS	'\0'

/*
 *
 */

#define	S_SHIFT_OCTAVE		12
#define	S_SHIFT_HIGHT(X,O)	((X)+(O)*S_SHIFT_OCTAVE)
#define	S_SHIFT_LOW(X,O)	((X)-(O)*S_SHIFT_OCTAVE)

/*
 *
 */

#define	S_MIDI___	0	/* 休 */

#define	S_MIDI_C4	0x3c	/* ド */
#define	S_MIDI_D4	0x3e	/* レ */
#define	S_MIDI_E4	0x40	/* ミ */
#define	S_MIDI_F4	0x41	/* ファ */
#define	S_MIDI_G4	0x43	/* ソ */
#define	S_MIDI_A4	0x45	/* ラ */
#define	S_MIDI_B4	0x47	/* シ */
#define	S_MIDI_C5	0x48	/* ド */

/*
 *
 */

#define	S_MIDI_LEN_8	500
#define	S_MIDI_LEN_4	(S_MIDI_LEN_8*2)
#define	S_MIDI_LEN_2	(S_MIDI_LEN_4*2)
#define	S_MIDI_LEN_1	(S_MIDI_LEN_2*2)

/*
 *
 */

#define	Sleep(S)	usleep((S)*1000)

/*
 *
 */

static int S_midi_init_flag = 0;
static int S_midi_hMidiOut; /* MIDI 出力デバイスのハンドル*/

/*
 *
 */

static void S_midi_note_on ( int note ) {
  unsigned char data[3] = {0x90, 60, 127};

  if ( S_midi_init_flag == 1 ) {
	data[1] = note;
	write(S_midi_hMidiOut, data, sizeof(data));
  }
}


static void S_midi_note_off ( int note ) {
  unsigned char data[3] = {0x80, 60, 127};

  if ( S_midi_init_flag == 1 ) {
	data[1] = note;
	write(S_midi_hMidiOut, data, sizeof(data));
  }
}

/*
 *
 */

static void S_midi_atexit() {
  int i;

  if ( S_midi_init_flag == 1 ) {
	for ( i = S_MIDI_C4; i <= S_MIDI_C5; i++ ) {
	  S_midi_note_off ( i );
	}
	close ( S_midi_hMidiOut );
	S_midi_init_flag = 0;
  }
}

/*
 *
 */

static void S_midi_init() {
  char* device =  "/dev/snd/midiC1D0" ;

  if ( S_midi_init_flag == 0 ) {
	S_midi_init_flag = 1;

	S_midi_hMidiOut = open(device, O_WRONLY, 0);
	if ( S_midi_hMidiOut < 0) {
      printf("Error: cannot open %s\n", device);
      exit(1);
	}

	atexit ( S_midi_atexit );
  }

}

/*
 *
 */

static void s_midi_start ( int midi ) {

  S_midi_init();
  S_midi_note_on ( midi );
}

static void s_midi_stop ( int midi ) {

  S_midi_note_off ( midi );
}

static void s_midi_sleep ( int len ) {

	Sleep ( len );
}

/*
 *
 */

static int S_midi_length = S_MIDI_LEN_4;

static void s_midi_set_length ( int len ) {

	S_midi_length = len;
}

static void s_midi_wait ( void ) {

	s_midi_sleep ( S_midi_length );
}

/*
 *
 */

static void s_midi_play ( int midi ) {

  if ( midi == S_MIDI___ ) {
	s_midi_wait();
  } else {
	s_midi_start ( midi );
	s_midi_wait();
	s_midi_stop ( midi );
  }
}

/*
 *
 */

static void s_midi_harmony ( int midi1, int midi2, int midi3 ) {

  S_midi_init();

  s_midi_start ( midi1 );
  s_midi_start ( midi2 );
  s_midi_start ( midi3 );

  s_midi_wait();

  s_midi_stop ( midi1 );
  s_midi_stop ( midi2 );
  s_midi_stop ( midi3 );

}

/*
 *
 */

#define MAX_MODE 3

#define	MIDI_STOP	0
#define	MIDI_START	1

static char *s_midi_play_one_code_cotrol ( char *midi, int control ) {
  int code = S_MIDI___;
  int alpha = 'z';
  int number = 4;
  int flat = 0;
  int ch;
  int mode = 0;
  static char *tbl = "C D EF G A B";

  S_midi_init();

  for ( mode = 0; *midi != EOS && mode < MAX_MODE; mode++ ) {
	while ( isspace (ch = *midi++) ) {
	  /* skip space */
	}
	switch ( mode ) {
	case 0:
		if ( 'a' <= ch && ch <= 'g' ) {
		  ch = ch - 'a' + 'A';
		}
		if ( 'A' <= ch && ch <= 'G' ) {
		  alpha = ch;
		} else if ( ch == 'z' ) {
		  mode = MAX_MODE;
		} else {	/* illegal code : skip it */
		  return midi;
		}
		break;
	case 1:
	  if ( isdigit ( ch ) ) {
		number = ch;
	  } else {
		midi--;
		mode = MAX_MODE;
	  }
		break;
	case 2:
	  if ( ch == '+' ) {
		flat = 1;
	  } else if ( ch == '-' ) {
		flat = -1;
	  } else {
		midi--;
	  }
	  mode = MAX_MODE;
	  break;
	defaut:
		break;
	}
  }

  if ( alpha == 'z' ) {
	code = S_MIDI___;
  } else {
	code = (number-'0'+1)*S_SHIFT_OCTAVE;	/* C1 : 24, C2 : 36 */
	code += strchr ( tbl, alpha ) - tbl;
	code += flat;
  }

  //  printf ( "code = %d\n", code );

  if ( code != S_MIDI___ ) {
	if ( control == MIDI_START ) {
	  s_midi_start ( code );
	} else {
	  s_midi_stop ( code );
	}
  }

  return midi;
}

/*
 *
 */

static char *s_midi_play_one_code ( char *midi ) {

  s_midi_play_one_code_cotrol ( midi, MIDI_START );
  s_midi_wait();
  return s_midi_play_one_code_cotrol ( midi, MIDI_STOP );

}

/*
 *
 */

static void s_midi_play_code ( char *midi ) {
  char *p = midi;

  for ( p = midi; *p != EOS; ) {
	p = s_midi_play_one_code_cotrol ( p, MIDI_START );
  }

  s_midi_wait();

  for ( p = midi; *p != EOS; ) {
	p = s_midi_play_one_code_cotrol ( p, MIDI_STOP );
  }

}

/*
 *
 */

static int s_midi_default_tempo = S_MIDI_LEN_4;

static void s_midi_set_tempo ( int tempo ) {

  s_midi_default_tempo = tempo;

}

/*
 *
 */

#define UTF_8_CHAR_LENGTH	3		/* strlen ( "ド" ) == 3 (UTF-8) */
#define	CODE_LENGTH			2		/* strlen ( "C4" ) == 2 */

static void s_midi_play_doremi ( char *midi ) {
  static char *tbl = "ラシドレミフソ";		/* ABCDEFG */

  char code[UTF_8_CHAR_LENGTH+1];
  char note[CODE_LENGTH*10+1];
  int len;

  code[UTF_8_CHAR_LENGTH] = EOS;
  note[CODE_LENGTH] = EOS;

  while ( *midi != EOS ) {

	while ( isspace ( *midi ) ) {
	  midi++;
	}

	if ( *midi == EOS ) {
	  break;
	}

	len = 1;
	note[1] = '4';
	strncpy ( code, midi, UTF_8_CHAR_LENGTH );
	midi += UTF_8_CHAR_LENGTH;

	if ( !strcmp ( code, "休" ) || !strcmp ( code, "ッ" ) ) {
	  note[0] = 'z';
	  note[1] = EOS;
	} else {
	  int loop;
	  int i;

	  if ( !strncmp ( code, "「", UTF_8_CHAR_LENGTH ) ) {
		strncpy ( code, midi, UTF_8_CHAR_LENGTH );
		midi += UTF_8_CHAR_LENGTH;
		loop = 2;
	  } else {
		loop = 1;
	  }

	  i = 0;
	  while ( loop > 0 ) {

		if ( loop == 2 && !strncmp ( code, "」", UTF_8_CHAR_LENGTH ) ) {
		  note[i] = EOS;
		  break;
		}

		if ( strstr ( tbl, code ) != NULL ) {
		  note[i++] = 'A' + (strstr ( tbl, code ) - tbl) / UTF_8_CHAR_LENGTH;
		  note[i] = '4';
		}

		if ( !strncmp ( midi, "ァ", UTF_8_CHAR_LENGTH ) ) {
		  midi += UTF_8_CHAR_LENGTH;
		}

		while ( !strncmp ( midi, "’", UTF_8_CHAR_LENGTH ) ) {
		  midi += UTF_8_CHAR_LENGTH;
		  note[i] += 1;
		}

		while ( !strncmp ( midi, "、", UTF_8_CHAR_LENGTH ) ) {
		  midi += UTF_8_CHAR_LENGTH;
		  note[i] -= 1;
		}

		note[++i] = EOS;

		if ( loop == 2 ) {
		  strncpy ( code, midi, UTF_8_CHAR_LENGTH );
		  midi += UTF_8_CHAR_LENGTH;
		} else {
		  break;
		}
	  }
	}

	while (    !strncmp ( midi, "　", UTF_8_CHAR_LENGTH )
			|| !strncmp ( midi, "ー", UTF_8_CHAR_LENGTH ) ) {
	  len++;
	  midi += UTF_8_CHAR_LENGTH;
	  //	  printf ( "%s (%p)\n", midi, midi );
	}

	//	printf ( "%d : %s(%s)\n", len, code, note );

	s_midi_set_length ( s_midi_default_tempo * len );
	s_midi_play_code ( note );
  }

}

/*
 *
 *	cf. http://musicology.s19.xrea.com/x/reading.html
 */

static void s_midi_play_speed ( char *speed ) {

  if ( !strcmp ( speed, "Adagio" ) ) {	/* おそく */
	s_midi_set_tempo ( S_MIDI_LEN_2 );	
  } else if ( !strcmp ( speed, "おそく" ) ) {	/* おそく */
	s_midi_set_tempo ( S_MIDI_LEN_2 );
  } else if ( !strcmp ( speed, "Allegro" ) ) {	/* はやく */
	s_midi_set_tempo ( S_MIDI_LEN_8 );
  } else if ( !strcmp ( speed, "はやく" ) ) {	/* はやく */
	s_midi_set_tempo ( S_MIDI_LEN_8 );
  } else if ( !strcmp ( speed, "Moderato" ) ) {	/* ふつう */
	s_midi_set_tempo ( S_MIDI_LEN_4 );
  } else if ( !strcmp ( speed, "ふつう" ) ) {	/* ふつう */
	s_midi_set_tempo ( S_MIDI_LEN_4 );
  } else {
	s_midi_set_tempo ( S_MIDI_LEN_4 );
  }

}

/*
 *
 */

#endif

/*
 *
 */
